class ConstraintFilterMixin:
    def _constraint_filter(self, constraints, **kwargs):
        if type(constraints) not in (tuple, list):
            raise ClaripyValueError("The extra_constraints argument should be a list of constraints.")

        if len(constraints) == 0:
            return constraints

        filtered = super(ConstraintFilterMixin, self)._constraint_filter(constraints, **kwargs)
        ccs = [ self._concrete_constraint(c) for c in filtered ]
        if False in ccs:
            raise UnsatError("Constraints contain False.")
        else:
            return tuple((o if n is None else o) for o,n in zip(constraints, ccs) if n is not True)

    def add(self, constraints, **kwargs):
        try:
            ec = self._constraint_filter(constraints)
        except UnsatError:
            # filter out concrete False
            ec = list(c for c in constraints if c not in {False, false}) + [ false ]

        if len(constraints) == 0:
            return [ ]

        if len(ec) > 0:
            return super(ConstraintFilterMixin, self).add(ec, **kwargs)
        else:
            return [ ]

    def satisfiable(self, extra_constraints=(), **kwargs):
        try:
            ec = self._constraint_filter(extra_constraints)
            return super(ConstraintFilterMixin, self).satisfiable(extra_constraints=ec, **kwargs)
        except UnsatError:
            return False

    def eval(self, e, n, extra_constraints=(), **kwargs):
        ec = self._constraint_filter(extra_constraints)
        return super(ConstraintFilterMixin, self).eval(e, n, extra_constraints=ec, **kwargs)

    def batch_eval(self, exprs, n, extra_constraints=(), **kwargs):
        ec = self._constraint_filter(extra_constraints)
        return super(ConstraintFilterMixin, self).batch_eval(exprs, n, extra_constraints=ec, **kwargs)

    def max(self, e, extra_constraints=(), **kwargs):
        ec = self._constraint_filter(extra_constraints)
        return super(ConstraintFilterMixin, self).max(e, extra_constraints=ec, **kwargs)

    def min(self, e, extra_constraints=(), **kwargs):
        ec = self._constraint_filter(extra_constraints)
        return super(ConstraintFilterMixin, self).min(e, extra_constraints=ec, **kwargs)

    def solution(self, e, v, extra_constraints=(), **kwargs):
        ec = self._constraint_filter(extra_constraints)
        return super(ConstraintFilterMixin, self).solution(e, v, extra_constraints=ec, **kwargs)

    def is_true(self, e, extra_constraints=(), **kwargs):
        ec = self._constraint_filter(extra_constraints)
        return super(ConstraintFilterMixin, self).is_true(e, extra_constraints=ec, **kwargs)

    def is_false(self, e, extra_constraints=(), **kwargs):
        ec = self._constraint_filter(extra_constraints)
        return super(ConstraintFilterMixin, self).is_false(e, extra_constraints=ec, **kwargs)

from ..errors import UnsatError, ClaripyValueError
from .. import false



import decimal
import functools
import math
import struct
from decimal import Decimal
from enum import Enum

from .errors import ClaripyOperationError
from .backend_object import BackendObject

def compare_sorts(f):
    @functools.wraps(f)
    def compare_guard(self, o):
        if self.sort != o.sort:
            raise TypeError("FPVs are differently-sorted ({} and {})".format(self.sort, o.sort))
        return f(self, o)

    return compare_guard

def normalize_types(f):
    @functools.wraps(f)
    def normalize_helper(self, o):
        if isinstance(o, float):
            o = FPV(o, self.sort)

        if not isinstance(self, FPV) or not isinstance(o, FPV):
            raise TypeError("must have two FPVs")

        return f(self, o)

    return normalize_helper


class RM(Enum):
    # see https://en.wikipedia.org/wiki/IEEE_754#Rounding_rules
    RM_NearestTiesEven = 'RM_RNE'
    RM_NearestTiesAwayFromZero = 'RM_RNA'
    RM_TowardsZero = 'RM_RTZ'
    RM_TowardsPositiveInf = 'RM_RTP'
    RM_TowardsNegativeInf = 'RM_RTN'

    @staticmethod
    def default():
        return RM.RM_NearestTiesEven

    def pydecimal_equivalent_rounding_mode(self):
        return {
            RM.RM_TowardsPositiveInf:      decimal.ROUND_CEILING,
            RM.RM_TowardsNegativeInf:      decimal.ROUND_FLOOR,
            RM.RM_TowardsZero:             decimal.ROUND_DOWN,
            RM.RM_NearestTiesEven:         decimal.ROUND_HALF_EVEN,
            RM.RM_NearestTiesAwayFromZero: decimal.ROUND_UP,
        }[self]


RM_NearestTiesEven          = RM.RM_NearestTiesEven
RM_NearestTiesAwayFromZero  = RM.RM_NearestTiesAwayFromZero
RM_TowardsZero              = RM.RM_TowardsZero
RM_TowardsPositiveInf       = RM.RM_TowardsPositiveInf
RM_TowardsNegativeInf       = RM.RM_TowardsNegativeInf


class FSort:
    def __init__(self, name, exp, mantissa):
        self.name = name
        self.exp = exp
        self.mantissa = mantissa

    def __eq__(self, other):
        return self.exp == other.exp and self.mantissa == other.mantissa

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash((self.name, self.exp, self.mantissa))

    @property
    def length(self):
        return self.exp + self.mantissa

    @staticmethod
    def from_size(n):
        if n == 32:
            return FSORT_FLOAT
        elif n == 64:
            return FSORT_DOUBLE
        else:
            raise ClaripyOperationError('{} is not a valid FSort size'.format(n))

    @staticmethod
    def from_params(exp, mantissa):
        if exp == 8 and mantissa == 24:
            return FSORT_FLOAT
        elif exp == 11 and mantissa == 53:
            return FSORT_DOUBLE
        else:
            raise ClaripyOperationError("unrecognized FSort params")

FSORT_FLOAT = FSort('FLOAT', 8, 24)
FSORT_DOUBLE = FSort('DOUBLE', 11, 53)


class FPV(BackendObject):
    __slots__ = ['sort', 'value']

    def __init__(self, value, sort):
        if not isinstance(value, float) or sort not in {FSORT_FLOAT, FSORT_DOUBLE}:
            raise ClaripyOperationError("FPV needs a sort (FSORT_FLOAT or FSORT_DOUBLE) and a float value")

        self.value = value
        self.sort = sort

    def __hash__(self):
        return hash((self.value, self.sort))

    def __getstate__(self):
        return self.value, self.sort

    def __setstate__(self, st):
        self.value, self.sort = st

    def __abs__(self):
        return FPV(abs(self.value), self.sort)

    def __neg__(self):
        return FPV(-self.value, self.sort)

    def fpSqrt(self):
        return FPV(math.sqrt(self.value), self.sort)

    @normalize_types
    @compare_sorts
    def __add__(self, o):
        return FPV(self.value + o.value, self.sort)

    @normalize_types
    @compare_sorts
    def __sub__(self, o):
        return FPV(self.value - o.value, self.sort)

    @normalize_types
    @compare_sorts
    def __mul__(self, o):
        return FPV(self.value * o.value, self.sort)

    @normalize_types
    @compare_sorts
    def __mod__(self, o):
        return FPV(self.value % o.value, self.sort)

    @normalize_types
    @compare_sorts
    def __truediv__(self, o):
        try:
            return FPV(self.value / o.value, self.sort)
        except ZeroDivisionError:
            if str(self.value * o.value)[0] == '-':
                return FPV(float('-inf'), self.sort)
            else:
                return FPV(float('inf'), self.sort)

    def __floordiv__(self, other): # decline to involve integers in this floating point process
        return self.__truediv__(other)

    #
    # Reverse arithmetic stuff
    #

    @normalize_types
    @compare_sorts
    def __radd__(self, o):
        return FPV(o.value + self.value, self.sort)

    @normalize_types
    @compare_sorts
    def __rsub__(self, o):
        return FPV(o.value - self.value, self.sort)

    @normalize_types
    @compare_sorts
    def __rmul__(self, o):
        return FPV(o.value * self.value, self.sort)

    @normalize_types
    @compare_sorts
    def __rmod__(self, o):
        return FPV(o.value % self.value, self.sort)

    @normalize_types
    @compare_sorts
    def __rtruediv__(self, o):
        try:
            return FPV(o.value / self.value, self.sort)
        except ZeroDivisionError:
            if str(o.value * self.value)[0] == '-':
                return FPV(float('-inf'), self.sort)
            else:
                return FPV(float('inf'), self.sort)

    def __rfloordiv__(self, other): # decline to involve integers in this floating point process
        return self.__rtruediv__(other)

    #
    # Boolean stuff
    #

    @normalize_types
    @compare_sorts
    def __eq__(self, o):
        return self.value == o.value

    @normalize_types
    @compare_sorts
    def __ne__(self, o):
        return self.value != o.value

    @normalize_types
    @compare_sorts
    def __lt__(self, o):
        return self.value < o.value

    @normalize_types
    @compare_sorts
    def __gt__(self, o):
        return self.value > o.value

    @normalize_types
    @compare_sorts
    def __le__(self, o):
        return self.value <= o.value

    @normalize_types
    @compare_sorts
    def __ge__(self, o):
        return self.value >= o.value

    def __repr__(self):
        return 'FPV({:f}, {})'.format(self.value, self.sort)

def fpToFP(a1, a2, a3=None):
    """
    Returns a FP AST and has three signatures:

        fpToFP(ubvv, sort)
            Returns a FP AST whose value is the same as the unsigned BVV `a1`
            and whose sort is `a2`.

        fpToFP(rm, fpv, sort)
            Returns a FP AST whose value is the same as the floating point `a2`
            and whose sort is `a3`.

        fpToTP(rm, sbvv, sort)
            Returns a FP AST whose value is the same as the signed BVV `a2` and
            whose sort is `a3`.
    """
    if isinstance(a1, BVV) and isinstance(a2, FSort):
        sort = a2
        if sort == FSORT_FLOAT:
            pack, unpack = 'I', 'f'
        elif sort == FSORT_DOUBLE:
            pack, unpack = 'Q', 'd'
        else:
            raise ClaripyOperationError("unrecognized float sort")

        try:
            packed = struct.pack('<' + pack, a1.value)
            unpacked, = struct.unpack('<' + unpack, packed)
        except OverflowError as e:
            # struct.pack sometimes overflows
            raise ClaripyOperationError("OverflowError: " + str(e)) from e

        return FPV(unpacked, sort)
    elif isinstance(a1, RM) and isinstance(a2, FPV) and isinstance(a3, FSort):
        return FPV(a2.value, a3)
    elif isinstance(a1, RM) and isinstance(a2, BVV) and isinstance(a3, FSort):
        return FPV(float(a2.signed), a3)
    else:
        raise ClaripyOperationError("unknown types passed to fpToFP")

def fpToFPUnsigned(_rm, thing, sort):
    """
    Returns a FP AST whose value is the same as the unsigned BVV `thing` and
    whose sort is `sort`.
    """
    # thing is a BVV
    return FPV(float(thing.value), sort)

def fpToIEEEBV(fpv):
    """
    Interprets the bit-pattern of the IEEE754 floating point number `fpv` as a
    bitvector.

    :return:    A BV AST whose bit-pattern is the same as `fpv`
    """
    if fpv.sort == FSORT_FLOAT:
        pack, unpack = 'f', 'I'
    elif fpv.sort == FSORT_DOUBLE:
        pack, unpack = 'd', 'Q'
    else:
        raise ClaripyOperationError("unrecognized float sort")

    try:
        packed = struct.pack('<' + pack, fpv.value)
        unpacked, = struct.unpack('<' + unpack, packed)
    except OverflowError as e:
        # struct.pack sometimes overflows
        raise ClaripyOperationError("OverflowError: " + str(e)) from e

    return BVV(unpacked, fpv.sort.length)

def fpFP(sgn, exp, mantissa):
    """
    Concatenates the bitvectors `sgn`, `exp` and `mantissa` and returns the
    corresponding IEEE754 floating point number.

    :return:    A FP AST whose bit-pattern is the same as the concatenated
                bitvector
    """
    concatted = Concat(sgn, exp, mantissa)
    sort = FSort.from_size(concatted.size())

    if sort == FSORT_FLOAT:
        pack, unpack = 'I', 'f'
    elif sort == FSORT_DOUBLE:
        pack, unpack = 'Q', 'd'
    else:
        raise ClaripyOperationError("unrecognized float sort")

    try:
        packed = struct.pack('<' + pack, concatted.value)
        unpacked, = struct.unpack('<' + unpack, packed)
    except OverflowError as e:
        # struct.pack sometimes overflows
        raise ClaripyOperationError("OverflowError: " + str(e)) from e

    return FPV(unpacked, sort)

def fpToSBV(rm, fp, size):
    try:
        rounding_mode = rm.pydecimal_equivalent_rounding_mode()
        val = int(Decimal(fp.value).to_integral_value(rounding_mode))
        return BVV(val, size)

    except (ValueError, OverflowError):
        return BVV(0, size)
    except Exception as ex:
        print("Unhandled error during floating point rounding! {}".format(ex))
        raise

def fpToUBV(rm, fp, size):
    # todo: actually make unsigned
    try:
        rounding_mode = rm.pydecimal_equivalent_rounding_mode()
        val = int(Decimal(fp.value).to_integral_value(rounding_mode))
        assert val & ((1 << size) - 1) == val, \
            f"Rounding produced values outside the BV range! rounding {fp.value} with rounding mode {rm} produced {val}"
        if val < 0:
            val = (1 << size) + val
        return BVV(val, size)

    except (ValueError, OverflowError):
        return BVV(0, size)

def fpEQ(a, b):
    """
    Checks if floating point `a` is equal to floating point `b`.
    """
    return a == b

def fpNE(a, b):
    """
    Checks if floating point `a` is not equal to floating point `b`.
    """
    return a != b

def fpGT(a, b):
    """
    Checks if floating point `a` is greater than floating point `b`.
    """
    return a > b

def fpGEQ(a, b):
    """
    Checks if floating point `a` is greater than or equal to floating point `b`.
    """
    return a >= b

def fpLT(a, b):
    """
    Checks if floating point `a` is less than floating point `b`.
    """
    return a < b

def fpLEQ(a, b):
    """
    Checks if floating point `a` is less than or equal to floating point `b`.
    """
    return a <= b

def fpAbs(x):
    """
    Returns the absolute value of the floating point `x`. So:

        a = FPV(-3.2, FSORT_DOUBLE)
        b = fpAbs(a)
        b is FPV(3.2, FSORT_DOUBLE)
    """
    return abs(x)

def fpNeg(x):
    """
    Returns the additive inverse of the floating point `x`. So:

        a = FPV(3.2, FSORT_DOUBLE)
        b = fpAbs(a)
        b is FPV(-3.2, FSORT_DOUBLE)
    """
    return -x

def fpSub(_rm, a, b):
    """
    Returns the subtraction of the floating point `a` by the floating point `b`.
    """
    return a - b

def fpAdd(_rm, a, b):
    """
    Returns the addition of two floating point numbers, `a` and `b`.
    """
    return a + b

def fpMul(_rm, a, b):
    """
    Returns the multiplication of two floating point numbers, `a` and `b`.
    """
    return a * b

def fpDiv(_rm, a, b):
    """
    Returns the division of the floating point `a` by the floating point `b`.
    """
    return a / b

def fpIsNaN(x):
    """
    Checks whether the argument is a floating point NaN.
    """
    return math.isnan(x)

def fpIsInf(x):
    """
    Checks whether the argument is a floating point infinity.
    """
    return math.isinf(x)

from .bv import BVV, Concat



class EvalStringsToASTsMixin(object):
    def eval_to_ast(self, e, n, extra_constraints=(), exact=None):
        if type(e) is String:
            return [StringV(v, ) for v in self.eval(e, n, extra_constraints=extra_constraints, exact=exact) ]

        return super(EvalStringsToASTsMixin, self).eval_to_ast(e, n, extra_constraints=extra_constraints, exact=None)

from .. import String, StringV



import logging

from ..ast.base import Base, _make_name, ASTCacheKey

l = logging.getLogger('claripy.ast.bool')

_boolv_cache = dict()

# This is a hilarious hack to get around some sort of bug in z3's python bindings, where
# under some circumstances stuff gets destructed out of order
def cleanup():
    global _boolv_cache #pylint:disable=global-variable-not-assigned
    del _boolv_cache
import atexit
atexit.register(cleanup)

class Bool(Base):
    __slots__ = ()

    @staticmethod
    def _from_bool(like, val): #pylint:disable=unused-argument
        return BoolV(val)

    def is_true(self):
        """
        Returns True if 'self' can be easily determined to be True. Otherwise, return False. Note that the AST *might*
        still be True (i.e., if it were simplified via Z3), but it's hard to quickly tell that.
        """
        return is_true(self)

    def is_false(self):
        """
        Returns True if 'self' can be easily determined to be False. Otherwise, return False. Note that the AST *might*
        still be False (i.e., if it were simplified via Z3), but it's hard to quickly tell that.
        """
        return is_false(self)


def BoolS(name, explicit_name=None):
    """
    Creates a boolean symbol (i.e., a variable).

    :param name:            The name of the symbol
    :param explicit_name:   If False, an identifier is appended to the name to ensure uniqueness.

    :return:                A Bool object representing this symbol.
    """
    n = _make_name(name, -1, False if explicit_name is None else explicit_name)
    return Bool('BoolS', (n,), variables={n}, symbolic=True)

def BoolV(val):
    try:
        return _boolv_cache[(val)]
    except KeyError:
        result = Bool('BoolV', (val,))
        _boolv_cache[val] = result
        return result

#
# some standard ASTs
#

true = BoolV(True)
false = BoolV(False)

#
# Bound operations
#

from .. import operations

Bool.__eq__ = operations.op('__eq__', (Bool, Bool), Bool)
Bool.__ne__ = operations.op('__ne__', (Bool, Bool), Bool)
Bool.intersection = operations.op('intersection', (Bool, Bool), Bool)

#
# Unbound operations
#

def If(*args):
    # the coercion here is strange enough that we'll just implement it manually
    if len(args) != 3:
        raise ClaripyOperationError("invalid number of args passed to If")

    args = list(args)

    if isinstance(args[0], bool):
        args[0] = BoolV(args[0])

    ty = None
    if isinstance(args[1], Base):
        ty = type(args[1])
    elif isinstance(args[2], Base):
        ty = type(args[2])
    else:
        raise ClaripyTypeError("true/false clause of If must have bearable types")

    if isinstance(args[1], Bits) and isinstance(args[2], Bits) and args[1].length != args[2].length:
        raise ClaripyTypeError("sized arguments to If must have the same length")

    if not isinstance(args[1], ty):
        if hasattr(ty, '_from_' + type(args[1]).__name__):
            convert = getattr(ty, '_from_' + type(args[1]).__name__)
            args[1] = convert(args[2], args[1])
        else:
            raise ClaripyTypeError("can't convert {} to {}".format(type(args[1]), ty))
    if not isinstance(args[2], ty):
        if hasattr(ty, '_from_' + type(args[2]).__name__):
            convert = getattr(ty, '_from_' + type(args[2]).__name__)
            args[2] = convert(args[1], args[2])
        else:
            raise ClaripyTypeError("can't convert {} to {}".format(type(args[2]), ty))

    if is_true(args[0]):
        return args[1].append_annotations(args[0].annotations)
    elif is_false(args[0]):
        return args[2].append_annotations(args[0].annotations)

    if isinstance(args[1], Base) and args[1].op == 'If' and args[1].args[0] is args[0]:
        return If(args[0], args[1].args[1], args[2])
    if isinstance(args[1], Base) and args[1].op == 'If' and args[1].args[0] is Not(args[0]):
        return If(args[0], args[1].args[2], args[2])
    if isinstance(args[2], Base) and args[2].op == 'If' and args[2].args[0] is args[0]:
        return If(args[0], args[1], args[2].args[2])
    if isinstance(args[2], Base) and args[2].op == 'If' and args[2].args[0] is Not(args[0]):
        return If(args[0], args[1], args[2].args[1])

    if args[1] is args[2]:
        return args[1]
    if args[1] is true and args[2] is false:
        return args[0]
    if args[1] is false and args[2] is true:
        return ~args[0]

    if issubclass(ty, Bits):
        return ty('If', tuple(args), length=args[1].length)
    else:
        return ty('If', tuple(args))

And = operations.op('And', Bool, Bool, bound=False)
Or = operations.op('Or', Bool, Bool, bound=False)
Not = operations.op('Not', (Bool,), Bool, bound=False)

Bool.__invert__ = Not
Bool.__and__ = And
Bool.__rand__ = And
Bool.__or__ = Or
Bool.__ror__ = Or

def is_true(e, exact=None): #pylint:disable=unused-argument
    for b in backends._quick_backends:
        try: return b.is_true(e)
        except BackendError: pass

    l.debug("Unable to tell the truth-value of this expression")
    return False

def is_false(e, exact=None): #pylint:disable=unused-argument
    for b in backends._quick_backends:
        try: return b.is_false(e)
        except BackendError: pass

    l.debug("Unable to tell the truth-value of this expression")
    return False

#For large tables, ite_dict that uses a binary search tree instead of a "linear" search tree.
#This improves Z3 search capability (eliminating branches) and decreases recursion depth:
#linear search trees make Z3 error out on tables larger than a couple hundred elements.)
def ite_dict(i, d, default):

    """
    Return an expression of if-then-else trees which expresses a switch tree
    :param i: The variable which may take on multiple values affecting the final result
    :param d: A dict mapping possible values for i to values which the result could be
    :param default: A default value that the expression should take on if `i` matches none of the keys of `d`
    :return: An expression encoding the result of the above
    """
    i = i.ast if type(i) is ASTCacheKey else i

    #for small dicts fall back to the linear implementation
    if len(d) < 4:
        return ite_cases([ (i == c, v) for c,v in d.items() ], default)

    #otherwise, binary search.
    #Find the median:
    keys = list(d.keys())
    keys.sort()
    split_val = keys[len(keys)//2]

    #split the dictionary
    dictLow = {c:v for c,v in d.items() if c <= split_val}
    dictHigh = {c:v for c,v in d.items() if c > split_val}

    valLow = ite_dict(i, dictLow, default)
    valHigh = ite_dict(i, dictHigh, default)
    return If(i <= split_val, valLow, valHigh)

def ite_cases(cases, default):
    """
    Return an expression of if-then-else trees which expresses a series of alternatives

    :param cases: A list of tuples (c, v). `c` is the condition under which `v` should be the result of the expression
    :param default: A default value that the expression should take on if none of the `c` conditions are satisfied
    :return: An expression encoding the result of the above
    """
    sofar = default
    for c,v in reversed(list(cases)):
        if is_true(v == sofar):
            continue
        sofar = If(c, v, sofar)
    return sofar

def reverse_ite_cases(ast):
    """
    Given an expression created by `ite_cases`, produce the cases that generated it
    :param ast:
    :return:
    """
    queue = [(true, ast)]
    while queue:
        condition, ast = queue.pop(0)
        if ast.op == 'If':
            queue.append((And(condition, ast.args[0]), ast.args[1]))
            queue.append((And(condition, Not(ast.args[0])), ast.args[2]))
        else:
            yield condition, ast

def constraint_to_si(expr):
    """
    Convert a constraint to SI if possible.

    :param expr:
    :return:
    """

    satisfiable = True
    replace_list = [ ]

    satisfiable, replace_list = backends.vsa.constraint_to_si(expr)

    # Make sure the replace_list are all ast.bvs
    for i in range(len(replace_list)):  #pylint:disable=consider-using-enumerate
        ori, new = replace_list[i]
        if not isinstance(new, Base):
            new = BVS(new.name, new._bits, min=new._lower_bound, max=new._upper_bound, stride=new._stride, explicit_name=True)
            replace_list[i] = (ori, new)

    return satisfiable, replace_list

from ..backend_manager import backends
from ..errors import ClaripyOperationError, ClaripyTypeError, BackendError
from .bits import Bits
from .bv import BVS



#pylint:disable=redefined-outer-name
from typing import TYPE_CHECKING
# Mypy is severely confused by this delayed import trickery, but works if we just pretend that the import
# happens here already
if TYPE_CHECKING:
    from .bits import Bits
    from .bv import BV
    from .vs import VS
    from .fp import FP
    from .bool import Bool, true, false
    from .int import Int
    from .base import Base
    from .strings import String
    from .. import ops as all_operations
else:
    Bits = lambda *args, **kwargs: None
    BV = lambda *args, **kwargs: None
    VS = lambda *args, **kwargs: None
    FP = lambda *args, **kwargs: None
    Bool = lambda *args, **kwargs: None
    Int = lambda *args, **kwargs: None
    Base = lambda *args, **kwargs: None
    true = lambda *args, **kwargs: None
    false = lambda *args, **kwargs: None
    String = lambda *args, **kwargs: None
    all_operations = None


def _import():
    global Bits, BV, VS, FP, Bool, Int, Base, String, true, false, all_operations

    from .bits import Bits
    from .bv import BV
    from .vs import VS
    from .fp import FP
    from .bool import Bool, true, false
    from .int import Int
    from .base import Base
    from .strings import String
    from .. import ops as all_operations



import subprocess
import logging

import re
from . import SMTLibSolverBackend, PopenSolverProxy
from ...errors import MissingSolverError

log = logging.getLogger(__name__)

def get_version():
    try:
        version_string = subprocess.check_output(['abc', '--help']).decode('utf-8')
        #version_match = re.match('This is CVC4 version (.*)\n', version_string)

        #if not version_match:
        #    return False, None, "Found malformed version string: {}".format(version_string)

        return True, None, None # True, version_match.group(1), None

    except subprocess.CalledProcessError as ex:
        return False, None, "Not found, error: {}".format(ex)
    except OSError as ex:
        return False, None, "Not found, error: {}".format(ex)


IS_INSTALLED, VERSION, ERROR = get_version()

class ABCProxy(PopenSolverProxy):
    def __init__(self):
        self.installed = False
        p = None
        super(ABCProxy, self).__init__(p)

    def create_process(self):
        if not IS_INSTALLED:
            raise MissingSolverError('ABC not found! Please install ABC before using this backend')
        p = subprocess.Popen(['abc'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.installed = True
        return p

class SolverBackendABC(SMTLibSolverBackend):
    def solver(self, timeout=None, max_memory=None):
        """
        This function should return an instance of whatever object handles
        solving for this backend. For example, in Z3, this would be z3.Solver().
        """
        return ABCProxy()

from ... import backend_manager as backend_manager
backend_manager.backends._register_backend(SolverBackendABC(), 'smtlib_abc', False, False)



from .bits import Bits
from ..ast.base import _make_name

from .. import operations
from .bool import Bool
from .bv import BV, BVS, BVV

import math


class String(Bits):
    """
    Base class that represent the AST of a String object and implements all the operation useful to create and modify the AST.

    Do not instantiate this class directly, instead use StringS or StringV to construct a symbol or value, and then use
    operations to construct more complicated expressions.
    """

    __slots__ = ('string_length',)

    # Identifier used by composite solver in order to identify if a certain constraints contains
    # variables of type string... In this case cvc4 would handle the solving part.
    #
    # TODO: Find a smarter way to do this!
    STRING_TYPE_IDENTIFIER = 'STRING_'
    GENERATED_BVS_IDENTIFIER = 'BVS_'
    MAX_LENGTH = 10000

    def __init__(self, *args, **kwargs):
        str_len = kwargs['length']
        kwargs['length'] *= 8
        super().__init__(*args, **kwargs)
        self.string_length = str_len

    def __getitem__(self, rng):
        '''
        This is a big endian indexer that takes its arguments as bits but returns bytes.
        Returns the range of bits in self defined by: [rng.start, rng.end], indexed from the end
        The bit range may not internally divide any bytes; i.e. low and (high+1) must both be divisible by 8
        Examples:
            self[7:0]  -- returns the last byte of self
            self[15:0] -- returns the last two bytes of self
            self[8:0]  -- Error! [8:0] is 9 bits, it asks for individual bits of the second to last byte!
            self[8:1]  -- Error! [8:1] asks for 1 bit from the second to last byte and 7 from the last byte!
        '''
        if type(rng) is slice:
            bits_low = rng.start if rng.start is not None else 0
            bits_high = rng.stop if rng.stop is not None else 8*(self.string_length - 1)
            if bits_high % 8 != 0 or (bits_low+1) % 8 != 0:
                raise ValueError('Bit indicies must not internally divide bytes!')
            # high / low form a reverse-indexed byte index
            high = bits_high // 8
            low = bits_low // 8
            if high < 0:
                high = self.string_length + high
            if low < 0:
                low = self.string_length + low
            # StrSubstr takes a front-indexed byte index as a starting point, and a length
            start_idx = self.string_length - 1 - low
            if high > low:
                return StrSubstr(start_idx, 0, self)
            return StrSubstr(start_idx, 1 + low - high, self)
        else:
            raise ValueError("Only slices allowed for string extraction")

    @staticmethod
    # TODO: Figure out what to convert here
    #
    # The use case that i found useful here is during the concretization
    # of a symbolic address in memory.
    # When angr has to do this it creates a claripy.If constraints in this form
    # If(condition, <value_read_if_condition_true>, <value_read_if_condition_false>).
    # In some case <value_read_if_condition_false> can be MEM_UNINITIALIZED expressed as BVV
    #
    # What should we return in this case?
    def _from_BV(like, value): # pylint: disable=unused-argument
        return value

    @staticmethod
    def _from_str(like, value): # pylint: disable=unused-argument
        return StringV(value)

    def strReplace(self, str_to_replace, replacement):
        """
        Replace the first occurence of str_to_replace with replacement

        :param claripy.ast.String str_to_replace: pattern that has to be replaced
        :param claripy.ast.String replacement: replacement pattern
        """
        return StrReplace(self, str_to_replace, replacement)

    def toInt(self, bitlength):
        """
        Convert the string to a bitvector holding the integer
        representation of the string

        :param bitlength: size of the biitvector holding the result
        """
        return StrToInt(self, bitlength)

    def indexOf(self, pattern, start_idx, bitlength):
        """
        Return the start index of the pattern inside the input string in a
        Bitvector representation, otherwise it returns -1 (always using a BitVector)

        :param bitlength: size of the biitvector holding the result
        """
        return StrIndexOf(self, pattern, start_idx, bitlength)

    def raw_to_bv(self):
        """
        A counterpart to FP.raw_to_bv - does nothing and returns itself.
        """
        if self.symbolic:
            return BVS(next(iter(self.variables)).replace(self.STRING_TYPE_IDENTIFIER, self.GENERATED_BVS_IDENTIFIER), self.length)
        else:
            return BVV(ord(self.args[0]), self.length)

    def raw_to_fp(self):
        return self.raw_to_bv().raw_to_fp()


def StringS(name, size, uninitialized=False, explicit_name=False, **kwargs):
    """
    Create a new symbolic string (analogous to z3.String())

    :param name:                 The name of the symbolic string (i. e. the name of the variable)
    :param size:                 The size in bytes of the string (i. e. the length of the string)
    :param uninitialized:        Whether this value should be counted as an "uninitialized" value in the course of an
                                 analysis.
    :param bool explicit_name:   If False, an identifier is appended to the name to ensure uniqueness.

    :returns:                    The String object representing the symbolic string
    """
    n = _make_name(String.STRING_TYPE_IDENTIFIER + name, size, False if explicit_name is None else explicit_name)
    result = String("StringS", (n, uninitialized), length=size, symbolic=True, eager_backends=None, uninitialized=uninitialized, variables={n}, **kwargs)
    return result

def StringV(value, length=None, **kwargs):
    """
    Create a new Concrete string (analogous to z3.StringVal())

    :param value: The constant value of the concrete string

    :returns:                    The String object representing the concrete string
    """

    if length is None:
        length = len(value)

    if length < len(value):
        raise ValueError("Can't make a concrete string value longer than the specified length!")

    result = String("StringV", (value, len(value)), length=length, **kwargs)
    return result

StrConcat = operations.op('StrConcat', String, String, calc_length=operations.str_concat_length_calc, bound=False)
StrSubstr = operations.op('StrSubstr', (BV, BV, String),
                        String, calc_length=operations.substr_length_calc, bound=False)
StrLen = operations.op('StrLen', (String, int), BV, calc_length=operations.strlen_bv_size_calc, bound=False)
StrReplace = operations.op('StrReplace', (String, String, String), String,
                        extra_check=operations.str_replace_check,
                        calc_length=operations.str_replace_length_calc, bound=False)
StrContains = operations.op("StrContains", (String, String), Bool, bound=False)
StrPrefixOf = operations.op("StrPrefixOf", (String, String), Bool, bound=False)
StrSuffixOf = operations.op("StrSuffixOf", (String, String), Bool, bound=False)
StrIndexOf = operations.op("StrIndexOf", (String, String, BV, int), BV, calc_length=operations.strindexof_bv_size_calc, bound=False)
StrToInt = operations.op("StrToInt", (String, int), BV, calc_length=operations.strtoint_bv_size_calc, bound=False)
IntToStr = operations.op("IntToStr", (BV,), String, calc_length=operations.int_to_str_length_calc, bound=False)
StrIsDigit = operations.op("StrIsDigit", (String,), Bool, bound=False)

# Equality / inequality check
String.__eq__ = operations.op('__eq__', (String, String), Bool)
String.__ne__ = operations.op('__ne__', (String, String), Bool)

# String manipulation
String.__add__ = StrConcat
String.StrSubstr = staticmethod(StrSubstr)
String.StrConcat = staticmethod(StrConcat)
String.StrLen = staticmethod(StrLen)
String.StrReplace = staticmethod(StrReplace)
String.StrContains = staticmethod(StrContains)
String.StrPrefixOf = staticmethod(StrPrefixOf)
String.StrSuffixOf = staticmethod(StrSuffixOf)
String.StrIndexOf = staticmethod(StrIndexOf)
String.StrToInt = staticmethod(StrToInt)
String.StrIsDigit = staticmethod(StrIsDigit)
String.IntToStr = staticmethod(IntToStr)



class SimplifyHelperMixin:
    def max(self, *args, **kwargs):
        self.simplify()
        return super(SimplifyHelperMixin, self).max(*args, **kwargs)

    def min(self, *args, **kwargs):
        self.simplify()
        return super(SimplifyHelperMixin, self).min(*args, **kwargs)

    def eval(self, e, n, *args, **kwargs):
        if n > 1:
            self.simplify()
        return super(SimplifyHelperMixin, self).eval(e, n, *args, **kwargs)

    def batch_eval(self, e, n, *args, **kwargs):
        if n > 1:
            self.simplify()
        return super(SimplifyHelperMixin, self).batch_eval(e, n, *args, **kwargs)
